from casadi import *

def easycall(fcn, *args):
    return fcn(args)[0] if len(args)==1 else fcn(args)

T = 10. # Time horizon
N = 20 # number of control intervals
M = 4 # RK4 steps per interval
DT = T/N/M

# Declare model variables
x1 = MX.sym('x1')
x2 = MX.sym('x2')
x = vertcat([x1, x2])
u = MX.sym('u')

# Model equations
xdot = vertcat([(1-x2**2)*x1 - x2 + u, x1])

# Objective term
L = x1**2 + x2**2 + u**2;

# Continuous time dynamics
f = Function('f', [x, u], [xdot, L])
X0 = MX.sym('X0', 2)
U = MX.sym('U')
X = X0
Q = 0
for j in range(M):
    k1, k1_q = easycall(f, X, U)
    k2, k2_q = easycall(f, X + DT/2 * k1, U)
    k3, k3_q = easycall(f, X + DT/2 * k2, U)
    k4, k4_q = easycall(f, X + DT * k3, U)
    X=X+DT/6*(k1 +2*k2 +2*k3 +k4)
    Q = Q + DT/6*(k1_q + 2*k2_q + 2*k3_q + k4_q)
F = Function('F', [X0, U], [X, Q])

# Evaluate at a test point
X_test, Q_test = easycall(F, [0.2, 0.3], 0.4)
print 'X_test = ', X_test
print 'Q_test = ', Q_test

# Start with an empty NLP
w=[]
w0 = []
lbw = []
ubw = []
J = 0
g=[]
lbg = []
ubg = []

# Formulate the NLP
Xk = MX([0, 1])
for k in range(N):
    # New NLP variable for the control
    Uk = MX.sym('U_' + str(k))
    w += [Uk]
    lbw += [-1]
    ubw += [1]
    w0 += [0]

    # Integrate till the end of the interval
    Xk, Jk = easycall(F, Xk, Uk)
    J=J+Jk

    # Add inequality constraint
    g += [Xk[0]]
    lbg += [-.25]
    ubg += [inf]

# Create an NLP solver
prob = {'f': J, 'x': vertcat(w), 'g': vertcat(g)}
solver = nlpsol('solver', 'ipopt', prob);

# Solve the NLP
arg = {'x0': w0, 'lbx': lbw, 'ubx': ubw, \
             'lbg': lbg, 'ubg': ubg}
sol = solver(arg)
w_opt = sol['x']

# Plot the solution
u_opt = w_opt
x_opt = [[0, 1]]
for k in range(N):
    Xk, Jk = easycall(F, x_opt[-1], u_opt[k]);
    x_opt += [Xk.full()]
x1_opt = [r[0] for r in x_opt]
x2_opt = [r[1] for r in x_opt]

tgrid = [T/N*k for k in range(N+1)]
import matplotlib.pyplot as plt
plt.figure(1)
plt.clf()
plt.plot(tgrid, x1_opt, '--')
plt.plot(tgrid, x2_opt, '-')
plt.step(tgrid, vertcat((DM.nan(1), u_opt)), '-.')
plt.xlabel('t')
plt.legend(['x1','x2','u'])
plt.grid()
plt.show()
